# automatically generated by the FlatBuffers compiler, do not modify

# namespace: meta

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Generic **meta data attributes** for objects in other tables. Primary key of this table is ``(table_oid, object_oid, attribute)``.
class Attribute(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Attribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Attribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Table of the object holding the attribute.
    # Attribute
    def TableOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Attribute
    def TableOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Attribute
    def TableOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Attribute
    def TableOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Object (within the table) holding the attribute.
    # Attribute
    def ObjectOid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Attribute
    def ObjectOidAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Attribute
    def ObjectOidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Attribute
    def ObjectOidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Attribute name (part of the key).
    # Attribute
    def Attribute(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Timestamp when the attribute was last modified (or first created).
    # Attribute
    def Modified(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # CBOR-serialized, object-valued attribute.
    # Attribute
    def Value(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Attribute
    def ValueAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Attribute
    def ValueLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Attribute
    def ValueIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0


def AttributeStart(builder):
    builder.StartObject(5)


def Start(builder):
    AttributeStart(builder)


def AttributeAddTableOid(builder, tableOid):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(tableOid), 0)


def AddTableOid(builder, tableOid):
    AttributeAddTableOid(builder, tableOid)


def AttributeStartTableOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTableOidVector(builder, numElems: int) -> int:
    return AttributeStartTableOidVector(builder, numElems)


def AttributeAddObjectOid(builder, objectOid):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(objectOid), 0)


def AddObjectOid(builder, objectOid):
    AttributeAddObjectOid(builder, objectOid)


def AttributeStartObjectOidVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartObjectOidVector(builder, numElems: int) -> int:
    return AttributeStartObjectOidVector(builder, numElems)


def AttributeAddAttribute(builder, attribute):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(attribute), 0)


def AddAttribute(builder, attribute):
    AttributeAddAttribute(builder, attribute)


def AttributeAddModified(builder, modified):
    builder.PrependUint64Slot(3, modified, 0)


def AddModified(builder, modified):
    AttributeAddModified(builder, modified)


def AttributeAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0)


def AddValue(builder, value):
    AttributeAddValue(builder, value)


def AttributeStartValueVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartValueVector(builder, numElems: int) -> int:
    return AttributeStartValueVector(builder, numElems)


def AttributeEnd(builder):
    return builder.EndObject()


def End(builder):
    return AttributeEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from uuid import UUID

import numpy as np
from zlmdb import MapBytes20TimestampUuid, MapUuidFlatBuffers, flatbuffers, table

from cfxdb import pack_uint256, unpack_uint256
from cfxdb.gen.xbrmm import (
    Channel as ChannelGen,
)
from cfxdb.gen.xbrmm import (
    ChannelBalance as ChannelBalanceGen,
)
from cfxdb.gen.xbrmm import (
    ChannelState as ChannelStateGen,
)
from cfxdb.gen.xbrmm import (
    ChannelType as ChannelTypeGen,
)

ChannelType = ChannelTypeGen.ChannelType
ChannelState = ChannelStateGen.ChannelState


class _ChannelGen(ChannelGen.Channel):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsChannel(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _ChannelGen()
        x.Init(buf, n + offset)
        return x

    def MarketOidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def MemberOidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ChannelOidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def OpenAtAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def MarketmakerAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ActorAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def DelegateAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def RecipientAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def AmountAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ClosingAtAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ClosedAtAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def CloseMmSigAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def CloseDelSigAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def CloseBalanceAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(42))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ClosedTxAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(46))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Channel(object):
    """
    XBR off-chain payment or paying channel.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        self._market_oid = None
        self._member_oid = None
        self._channel_oid = None
        self._timestamp = None
        self._open_at = None
        self._seq = None
        self._channel_type = None
        self._marketmaker = None
        self._actor = None
        self._delegate = None
        self._recipient = None
        self._amount = None
        self._timeout = None
        self._state = None
        self._closing_at = None
        self._closed_at = None
        self._close_mm_sig = None
        self._close_del_sig = None
        self._close_channel_seq = None
        self._close_balance = None
        self._close_is_final = None
        self._closed_tx = None

    def marshal(self) -> dict:
        obj = {
            "market_oid": self.market_oid.bytes if self.market_oid else None,
            "member_oid": self.member_oid.bytes if self.member_oid else None,
            "channel_oid": self.channel_oid.bytes if self.channel_oid else None,
            "timestamp": int(self.timestamp) if self.timestamp else None,
            "open_at": pack_uint256(self.open_at) if self.open_at else None,
            "seq": self.seq,
            "channel_type": self.channel_type,
            "marketmaker": bytes(self.marketmaker) if self.marketmaker else None,
            "actor": bytes(self.actor) if self.actor else None,
            "delegate": bytes(self.delegate) if self.delegate else None,
            "recipient": bytes(self.recipient) if self.recipient else None,
            "amount": pack_uint256(self.amount) if self.amount else 0,
            "timeout": self.timeout,
            "state": self.state,
            "closing_at": pack_uint256(self.closing_at) if self.closing_at else None,
            "closed_at": pack_uint256(self.closed_at) if self.closed_at else None,
            "close_mm_sig": bytes(self.close_mm_sig) if self.close_mm_sig else None,
            "close_del_sig": bytes(self.close_del_sig) if self.close_del_sig else None,
            "close_channel_seq": self.close_channel_seq,
            "close_balance": pack_uint256(self.close_balance),
            "close_is_final": self.close_is_final,
            "closed_tx": bytes(self.closed_tx) if self.closed_tx else None,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def market_oid(self) -> UUID:
        """
        The unique OID of the market.
        """
        if self._market_oid is None and self._from_fbs:
            if self._from_fbs.MarketOidLength():
                _market_oid = self._from_fbs.MarketOidAsBytes()
                self._market_oid = UUID(bytes=bytes(_market_oid))
        return self._market_oid

    @market_oid.setter
    def market_oid(self, value: UUID):
        assert value is None or isinstance(value, UUID)
        self._market_oid = value

    @property
    def member_oid(self) -> UUID:
        """
        The unique OID of the member.
        """
        if self._member_oid is None and self._from_fbs:
            if self._from_fbs.MemberOidLength():
                _member_oid = self._from_fbs.MemberOidAsBytes()
                self._member_oid = UUID(bytes=bytes(_member_oid))
        return self._member_oid

    @member_oid.setter
    def member_oid(self, value: UUID):
        assert value is None or isinstance(value, UUID)
        self._member_oid = value

    @property
    def channel_oid(self) -> UUID:
        """
        The unique OID of the channel.
        """
        if self._channel_oid is None and self._from_fbs:
            if self._from_fbs.ChannelOidLength():
                _channel_oid = self._from_fbs.ChannelOidAsBytes()
                self._channel_oid = UUID(bytes=bytes(_channel_oid))
        return self._channel_oid

    @channel_oid.setter
    def channel_oid(self, value: UUID):
        assert value is None or isinstance(value, UUID)
        self._channel_oid = value

    @property
    def timestamp(self) -> np.datetime64:
        """
        Database transaction time (epoch time in ns) of insert or last update.
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def open_at(self) -> int:
        """
        Block number (on the blockchain) when the actor (originally) joined the market.
        """
        if self._open_at is None and self._from_fbs:
            if self._from_fbs.OpenAtLength():
                _open_at = self._from_fbs.OpenAtAsBytes()
                self._open_at = unpack_uint256(bytes(_open_at))
            else:
                self._open_at = 0
        return self._open_at

    @open_at.setter
    def open_at(self, value: int):
        assert value is None or type(value) == int
        self._open_at = value

    @property
    def seq(self) -> int:
        """
        Global channel sequence number.
        """
        if self._seq is None and self._from_fbs:
            self._seq = self._from_fbs.Seq()
        return self._seq or 0

    @seq.setter
    def seq(self, value: int):
        assert value is None or type(value) == int
        self._seq = value

    @property
    def channel_type(self) -> int:
        """
        Channel type: payment channel (from XBR consumer to XBR market maker) or paying channel (from XBR market maker to XBR provider).
        """
        if self._channel_type is None and self._from_fbs:
            self._channel_type = self._from_fbs.ChannelType()
        return self._channel_type

    @channel_type.setter
    def channel_type(self, value: int):
        assert type(value) == int
        self._channel_type = value

    @property
    def marketmaker(self) -> bytes:
        """
        ID of the payment channel.
        """
        if self._marketmaker is None and self._from_fbs:
            if self._from_fbs.MarketmakerLength():
                self._marketmaker = self._from_fbs.MarketmakerAsBytes()
        return self._marketmaker

    @marketmaker.setter
    def marketmaker(self, value: bytes):
        assert value is None or type(value) == bytes
        self._marketmaker = value

    @property
    def actor(self) -> bytes:
        """
        Ethereum address of the sender (either XBR Consumer or XBR Market).
        """
        if self._actor is None and self._from_fbs:
            if self._from_fbs.ActorLength():
                self._actor = self._from_fbs.ActorAsBytes()
        return self._actor

    @actor.setter
    def actor(self, value: bytes):
        assert value is None or type(value) == bytes
        self._actor = value

    @property
    def delegate(self) -> bytes:
        """
        Ethereum address of the sender delegate (either XBR Consumer delegate or XBR Market delegate == market maker)
        """
        if self._delegate is None and self._from_fbs:
            if self._from_fbs.DelegateLength():
                self._delegate = self._from_fbs.DelegateAsBytes()
        return self._delegate

    @delegate.setter
    def delegate(self, value: bytes):
        assert value is None or type(value) == bytes
        self._delegate = value

    @property
    def recipient(self) -> bytes:
        """
        Ethereum address of the recipient (either XBR Market or XBR Provider)
        """
        if self._recipient is None and self._from_fbs:
            if self._from_fbs.RecipientLength():
                self._recipient = self._from_fbs.RecipientAsBytes()
        return self._recipient

    @recipient.setter
    def recipient(self, value: bytes):
        assert value is None or type(value) == bytes
        self._recipient = value

    @property
    def amount(self) -> int:
        """
        Amount of XBR tokens initially deposited into the payment channel.
        """
        if self._amount is None and self._from_fbs:
            if self._from_fbs.AmountLength():
                _amount = self._from_fbs.AmountAsBytes()
                self._amount = unpack_uint256(bytes(_amount))
            else:
                self._amount = 0
        return self._amount

    @amount.setter
    def amount(self, value: int):
        assert value is None or type(value) == int
        self._amount = value

    @property
    def timeout(self) -> int:
        """
        Payment channel (non-cooperative) closed timeout in blocks (on the blockchain).
        """
        if self._timeout is None and self._from_fbs:
            self._timeout = self._from_fbs.Timeout()
        return self._timeout

    @timeout.setter
    def timeout(self, value: int):
        assert type(value) == int
        self._timeout = value

    @property
    def state(self) -> int:
        """
        Current state of payment channel.
        """
        if self._state is None and self._from_fbs:
            self._state = self._from_fbs.State()
        return self._state

    @state.setter
    def state(self, value: int):
        assert type(value) == int
        self._state = value

    @property
    def closing_at(self) -> int:
        """
        Block number (on the blockchain) when the payment channel will close (at the latest).
        """
        if self._closing_at is None and self._from_fbs:
            if self._from_fbs.ClosingAtLength():
                _closing_at = self._from_fbs.ClosingAtAsBytes()
                self._closing_at = unpack_uint256(bytes(_closing_at))
            else:
                self._closed_at = 0
        return self._closing_at

    @closing_at.setter
    def closing_at(self, value: int):
        assert value is None or type(value) == int
        self._closing_at = value

    @property
    def closed_at(self) -> int:
        """
        Block number (on the blockchain) when the payment channel was finally closed.
        """
        if self._closed_at is None and self._from_fbs:
            if self._from_fbs.ClosedAtLength():
                _closed_at = self._from_fbs.ClosedAtAsBytes()
                self._closed_at = unpack_uint256(bytes(_closed_at))
            else:
                self._closed_at = 0
        return self._closed_at

    @closed_at.setter
    def closed_at(self, value: int):
        assert value is None or type(value) == int
        self._closed_at = value

    @property
    def close_mm_sig(self) -> int:
        """
        Closing signature by market maker.
        """
        if self._close_mm_sig is None and self._from_fbs:
            if self._from_fbs.CloseMmSigLength():
                self._close_mm_sig = self._from_fbs.CloseMmSigAsBytes()
        return self._close_mm_sig

    @close_mm_sig.setter
    def close_mm_sig(self, value: int):
        assert value is None or type(value) == bytes
        self._close_mm_sig = value

    @property
    def close_del_sig(self) -> bytes:
        """
        Closing signature by (seller or buyer) delegate.
        """
        if self._close_del_sig is None and self._from_fbs:
            if self._from_fbs.CloseDelSigLength():
                self._close_del_sig = self._from_fbs.CloseDelSigAsBytes()
        return self._close_del_sig

    @close_del_sig.setter
    def close_del_sig(self, value: bytes):
        assert value is None or type(value) == bytes
        self._close_del_sig = value

    @property
    def close_channel_seq(self) -> int:
        """
        Last off-chain, closing transaction: channel transaction sequence number.
        """
        if self._close_channel_seq is None and self._from_fbs:
            self._close_channel_seq = self._from_fbs.CloseChannelSeq()
        return self._close_channel_seq

    @close_channel_seq.setter
    def close_channel_seq(self, value: int):
        assert type(value) == int
        self._close_channel_seq = value

    @property
    def close_balance(self) -> int:
        """
        Remaining (closing) channel balance (XBR).
        """
        if self._close_balance is None and self._from_fbs:
            if self._from_fbs.CloseBalanceLength():
                _close_balance = self._from_fbs.CloseBalanceAsBytes()
                if _close_balance is not None:
                    self._close_balance = unpack_uint256(bytes(_close_balance))
                else:
                    self._close_balance = 0
            else:
                self._close_balance = 0
        return self._close_balance

    @close_balance.setter
    def close_balance(self, value: int):
        assert value is None or type(value) == int
        self._close_balance = value

    @property
    def close_is_final(self) -> bool:
        """
        Flag indication if close is final (happens immediately without a channel timeout).
        """
        if self._close_is_final is None and self._from_fbs:
            self._close_is_final = self._from_fbs.CloseIsFinal() is True
        return self._close_is_final

    @close_is_final.setter
    def close_is_final(self, value: bool):
        assert value is None or type(value) == bool
        self._close_is_final = value

    @property
    def closed_tx(self) -> bytes:
        """
        When channel was finally closed on-chain, the Ethereum transaction ID.
        """
        if self._closed_tx is None and self._from_fbs:
            if self._from_fbs.ClosedTxLength():
                self._closed_tx = self._from_fbs.ClosedTxAsBytes()
        return self._closed_tx

    @closed_tx.setter
    def closed_tx(self, value: bytes):
        assert value is None or type(value) == bytes
        self._closed_tx = value

    @staticmethod
    def cast(buf):
        return Channel(_ChannelGen.GetRootAsChannel(buf, 0))

    def build(self, builder):
        market_oid = self.market_oid
        if market_oid:
            market_oid = builder.CreateString(market_oid.bytes)

        member_oid = self.member_oid
        if member_oid:
            member_oid = builder.CreateString(member_oid.bytes)

        channel_oid = self.channel_oid
        if channel_oid:
            channel_oid = builder.CreateString(channel_oid.bytes)

        open_at = self.open_at
        if open_at:
            open_at = builder.CreateString(pack_uint256(open_at))

        marketmaker = self.marketmaker
        if marketmaker:
            marketmaker = builder.CreateString(bytes(marketmaker))

        actor = self.actor
        if actor:
            actor = builder.CreateString(bytes(actor))

        delegate = self.delegate
        if delegate:
            delegate = builder.CreateString(bytes(delegate))

        recipient = self.recipient
        if recipient:
            recipient = builder.CreateString(bytes(recipient))

        amount = self.amount
        if amount:
            amount = builder.CreateString(pack_uint256(amount))

        closing_at = self.closing_at
        if closing_at:
            closing_at = builder.CreateString(pack_uint256(closing_at))

        closed_at = self.closed_at
        if closed_at:
            closed_at = builder.CreateString(pack_uint256(closed_at))

        close_mm_sig = self.close_mm_sig
        if close_mm_sig:
            close_mm_sig = builder.CreateString(bytes(close_mm_sig))

        close_del_sig = self.close_del_sig
        if close_del_sig:
            close_del_sig = builder.CreateString(bytes(close_del_sig))

        close_balance = self.close_balance
        if close_balance:
            close_balance = builder.CreateString(pack_uint256(close_balance))

        closed_tx = self.closed_tx
        if closed_tx:
            closed_tx = builder.CreateString(bytes(closed_tx))

        ChannelGen.ChannelStart(builder)

        if market_oid:
            ChannelGen.ChannelAddMarketOid(builder, market_oid)

        if member_oid:
            ChannelGen.ChannelAddMemberOid(builder, member_oid)

        if channel_oid:
            ChannelGen.ChannelAddChannelOid(builder, channel_oid)

        if self.timestamp:
            ChannelGen.ChannelAddTimestamp(builder, int(self.timestamp))

        if open_at:
            ChannelGen.ChannelAddOpenAt(builder, open_at)

        if self.seq:
            ChannelGen.ChannelAddSeq(builder, self.seq)

        if self.channel_type:
            ChannelGen.ChannelAddChannelType(builder, self.channel_type)

        if marketmaker:
            ChannelGen.ChannelAddMarketmaker(builder, marketmaker)

        if actor:
            ChannelGen.ChannelAddActor(builder, actor)

        if delegate:
            ChannelGen.ChannelAddDelegate(builder, delegate)

        if recipient:
            ChannelGen.ChannelAddRecipient(builder, recipient)

        if amount:
            ChannelGen.ChannelAddAmount(builder, amount)

        ChannelGen.ChannelAddTimeout(builder, self.timeout)

        if self.state:
            ChannelGen.ChannelAddState(builder, int(self.state))

        if closing_at:
            ChannelGen.ChannelAddClosingAt(builder, closing_at)

        if closed_at:
            ChannelGen.ChannelAddClosedAt(builder, closed_at)

        if close_mm_sig:
            ChannelGen.ChannelAddCloseMmSig(builder, close_mm_sig)

        if close_del_sig:
            ChannelGen.ChannelAddCloseDelSig(builder, close_del_sig)

        if self.close_channel_seq:
            ChannelGen.ChannelAddCloseChannelSeq(builder, self.close_channel_seq)

        if close_balance:
            ChannelGen.ChannelAddCloseBalance(builder, close_balance)

        if self.close_is_final:
            ChannelGen.ChannelAddCloseIsFinal(builder, self.close_is_final)

        if closed_tx:
            ChannelGen.ChannelAddClosedTx(builder, closed_tx)

        final = ChannelGen.ChannelEnd(builder)

        return final


@table("b3d01946-85ae-49f3-ad96-b78194eb82fe", build=Channel.build, cast=Channel.cast)
class PaymentChannels(MapUuidFlatBuffers):
    """
    XBR payment channels by ``channel_oid``.

    Map :class:`zlmdb.MapUuidFlatBuffers` from ``channel_oid`` to :class:`cfxdb.xbr.Channel`
    """


@table("cffd5253-72f8-41a9-8b76-5e6ff3654e67")
class IndexPaymentChannelByDelegate(MapBytes20TimestampUuid):
    """
    Index: ``(delegate_adr, created_timestamp) -> channel_oid``
    """


@table("c78e0113-3b40-42b1-a15f-0b478ccf0de4")
class IndexPaymentChannelByActor(MapBytes20TimestampUuid):
    """
    Index: ``(actor_adr, created_timestamp) -> channel_oid``
    """


@table("4e7e7c8d-db0d-4dea-8409-ac8f21ce1e10", build=Channel.build, cast=Channel.cast)
class PayingChannels(MapUuidFlatBuffers):
    """
    XBR paying channels by ``channel_oid``.

    Map :class:`zlmdb.MapUuidFlatBuffers` from ``channel_oid`` to :class:`cfxdb.xbr.Channel`
    """


@table("cee954be-fdb2-43cc-8891-529d6c7a0c3b")
class IndexPayingChannelByDelegate(MapBytes20TimestampUuid):
    """
    Index: ``(delegate_adr, created_timestamp) -> channel_oid``
    """


@table("655a9d5f-0bdf-4c2a-8102-208f6da4a566")
class IndexPayingChannelByRecipient(MapBytes20TimestampUuid):
    """
    Index: ``(recipient_adr, created_timestamp) -> channel_oid``
    """


class _ChannelBalanceGen(ChannelBalanceGen.ChannelBalance):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsChannelBalance(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _ChannelBalanceGen()
        x.Init(buf, n + offset)
        return x

    def RemainingAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def InflightAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class ChannelBalance(object):
    """
    XBR payment channel current (off-chain) balance. The sum of ``Balance.remaining`` and ``Balance.inflight`` equals ``Channel.amount``.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        self._remaining = None
        self._inflight = None
        self._seq = None

    @staticmethod
    def parse(data: dict):
        assert type(data) == dict

        obj = ChannelBalance()

        if "remaining" in data:
            remaining = data["remaining"]
            assert type(remaining) == bytes and len(remaining) == 32
            obj._remaining = unpack_uint256(remaining)

        if "inflight" in data:
            inflight = data["inflight"]
            assert type(inflight) == bytes and len(inflight) == 32
            obj._inflight = unpack_uint256(inflight)

        if "seq" in data:
            seq = data["seq"]
            assert type(seq) == int
            obj._seq = unpack_uint256(seq)

        return obj

    def marshal(self) -> dict:
        obj = {
            "remaining": pack_uint256(self.remaining) if self.remaining else 0,
            "inflight": pack_uint256(self.inflight) if self.inflight else 0,
            "seq": self.seq or 0,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def remaining(self) -> int:
        """
        Amount of XBR tokens currently remaining in the payment channel.
        """
        if self._remaining is None and self._from_fbs:
            if self._from_fbs.RemainingLength():
                _remaining = self._from_fbs.RemainingAsBytes()
                self._remaining = unpack_uint256(bytes(_remaining))
            else:
                self._remaining = 0
        return self._remaining

    @remaining.setter
    def remaining(self, value: int):
        assert value is None or type(value) == int
        self._remaining = value

    @property
    def inflight(self) -> int:
        """
        Amount of XBR tokens reserved to in-flight purchase transactions.
        """
        if self._inflight is None and self._from_fbs:
            if self._from_fbs.InflightLength():
                _inflight = self._from_fbs.InflightAsBytes()
                self._inflight = unpack_uint256(bytes(_inflight))
            else:
                self._inflight = 0
        return self._inflight

    @inflight.setter
    def inflight(self, value: int):
        assert value is None or type(value) == int
        self._inflight = value

    @property
    def seq(self) -> int:
        """
        Sequence number of transactions on this balance starting from 0 when the payment channel is created.
        """
        if self._seq is None and self._from_fbs:
            self._seq = self._from_fbs.Seq()
        return self._seq or 0

    @seq.setter
    def seq(self, value: int):
        assert value is None or type(value) == int
        self._seq = value

    @staticmethod
    def cast(buf):
        return ChannelBalance(_ChannelBalanceGen.GetRootAsChannelBalance(buf, 0))

    def build(self, builder):
        remaining = self.remaining
        if remaining:
            remaining = builder.CreateString(pack_uint256(remaining))

        inflight = self.inflight
        if inflight:
            inflight = builder.CreateString(pack_uint256(inflight))

        ChannelBalanceGen.ChannelBalanceStart(builder)

        if remaining:
            ChannelBalanceGen.ChannelBalanceAddRemaining(builder, remaining)

        if inflight:
            ChannelBalanceGen.ChannelBalanceAddInflight(builder, inflight)

        if self.seq:
            ChannelBalanceGen.ChannelBalanceAddSeq(builder, self.seq)

        final = ChannelBalanceGen.ChannelBalanceEnd(builder)

        return final


@table("878ac002-a830-488b-bfe9-f06371b8eecb", build=ChannelBalance.build, cast=ChannelBalance.cast)
class PaymentChannelBalances(MapUuidFlatBuffers):
    """
    XBR payment channels current balances by ``channel_oid``.

    Map :class:`zlmdb.MapBytes20FlatBuffers` from ``channel_oid`` to :class:`cfxdb.xbr.Balance`
    """


@table("c0931d5d-6d5d-4f9c-b2a3-29664a0f4c07", build=ChannelBalance.build, cast=ChannelBalance.cast)
class PayingChannelBalances(MapUuidFlatBuffers):
    """
    XBR paying channels current balances by ``channel_oid``.

    Map :class:`zlmdb.MapBytes20FlatBuffers` from ``channel_oid`` to :class:`cfxdb.xbr.Balance`
    """



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import random
import timeit
import uuid

import numpy as np
import pytest
from autobahn import util
from autobahn.wamp.types import TransportDetails
from txaio import time_ns, with_twisted  # noqa
from zlmdb import flatbuffers

from cfxdb.realmstore import Session

DATA1 = {
    "authextra": {
        "transport": {
            "channel_framing": "websocket",
            "channel_id": {},
            "channel_serializer": None,
            "channel_type": "tcp",
            "http_cbtid": "ch0oFqC4EQMCqpYmj/78bQ5D",
            "http_headers_received": {
                "cache-control": "no-cache",
                "connection": "Upgrade",
                "host": "localhost:8080",
                "pragma": "no-cache",
                "sec-websocket-extensions": "permessage-deflate; client_no_context_takeover; client_max_window_bits",
                "sec-websocket-key": "FG9K1Vx44MqEE9c37YgPEw==",
                "sec-websocket-protocol": "wamp.2.json",
                "sec-websocket-version": "13",
                "upgrade": "WebSocket",
                "user-agent": "AutobahnPython/22.4.1.dev7",
            },
            "http_headers_sent": {"Set-Cookie": "cbtid=ch0oFqC4EQMCqpYmj/78bQ5D;max-age=604800"},
            "is_secure": False,
            "is_server": True,
            "own": None,
            "own_fd": -1,
            "own_pid": 28806,
            "own_tid": 28806,
            "peer": "tcp4:127.0.0.1:48812",
            "peer_cert": None,
            "websocket_extensions_in_use": [
                {
                    "client_max_window_bits": 13,
                    "client_no_context_takeover": False,
                    "extension": "permessage-deflate",
                    "is_server": True,
                    "mem_level": 5,
                    "server_max_window_bits": 13,
                    "server_no_context_takeover": False,
                }
            ],
            "websocket_protocol": "wamp.2.json",
        },
        "x_cb_node": "intel-nuci7-28788",
        "x_cb_peer": "unix",
        "x_cb_pid": 28797,
        "x_cb_worker": "test_router1",
    },
    "authid": "client1",
    "authmethod": "anonymous-proxy",
    "authprovider": "static",
    "authrole": "frontend",
    "session": 941369063710961,
    "transport": {
        "channel_framing": "rawsocket",
        "channel_id": {},
        "channel_serializer": "cbor",
        "channel_type": "tcp",
        "http_cbtid": None,
        "http_headers_received": None,
        "http_headers_sent": None,
        "is_secure": False,
        "is_server": None,
        "own": None,
        "own_fd": -1,
        "own_pid": 28797,
        "own_tid": 28797,
        "peer": "unix",
        "peer_cert": None,
        "websocket_extensions_in_use": None,
        "websocket_protocol": "wamp.2.cbor",
    },
}

DATA2 = {
    "channel_framing": "rawsocket",
    "channel_id": {},
    "channel_serializer": "cbor",
    "channel_type": "tcp",
    "http_cbtid": None,
    "http_headers_received": None,
    "http_headers_sent": None,
    "is_secure": False,
    "is_server": None,
    "own": None,
    "own_fd": -1,
    "own_pid": 14017,
    "own_tid": 14017,
    "peer": "unix",
    "peer_cert": None,
    "websocket_extensions_in_use": None,
    "websocket_protocol": "wamp.2.cbor",
}


def fill_session(session):
    _td1 = TransportDetails.parse(DATA1["transport"])
    _td2 = TransportDetails.parse(DATA1["authextra"]["transport"])
    _td3 = TransportDetails.parse(DATA2)

    session.arealm_oid = uuid.uuid4()
    session.oid = uuid.uuid4()
    session.session = util.id()
    session.joined_at = np.datetime64(time_ns() - 723 * 10**9, "ns")
    session.left_at = np.datetime64(time_ns(), "ns")
    session.node_oid = uuid.uuid4()
    session.node_authid = "intel-nuci7"
    session.worker_name = "router1"
    session.worker_pid = 28797
    session.transport = _td1.marshal()
    session.proxy_node_oid = uuid.uuid4()
    session.proxy_node_authid = "intel-nuci7"
    session.proxy_worker_name = "proxy1"
    session.proxy_worker_pid = 30992
    session.proxy_transport = _td3.marshal()
    session.realm = "realm-{}".format(uuid.uuid4())
    session.authid = util.generate_serial_number()
    session.authrole = random.choice(["admin", "user*", "guest", "anon*"])
    session.authmethod = random.choice(["wampcra", "cookie", "anonymous-proxy"])
    session.authprovider = random.choice(["static", "dynamic"])
    session.authextra = {
        "transport": _td2.marshal(),
        "x_cb_node": DATA1["authextra"].get("x_cb_node", None),
        "x_cb_peer": DATA1["authextra"].get("x_cb_peer", None),
        "x_cb_pid": DATA1["authextra"].get("x_cb_pid", None),
        "x_cb_worker": DATA1["authextra"].get("x_cb_worker", None),
    }


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


@pytest.fixture(scope="function")
def session():
    _session = Session()
    fill_session(_session)
    return _session


def test_session_roundtrip(session, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = session.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) in [1944, 1952]

    # create python object from bytes (flatbuffers)
    _session = Session.cast(data)

    assert _session.arealm_oid == session.arealm_oid
    assert _session.oid == session.oid
    assert _session.session == session.session
    assert _session.joined_at == session.joined_at
    assert _session.left_at == session.left_at
    assert _session.node_oid == session.node_oid
    assert _session.node_authid == session.node_authid
    assert _session.worker_name == session.worker_name
    assert _session.worker_pid == session.worker_pid
    assert _session.transport == session.transport
    assert _session.proxy_node_oid == session.proxy_node_oid
    assert _session.proxy_node_authid == session.proxy_node_authid
    assert _session.proxy_worker_name == session.proxy_worker_name
    assert _session.proxy_worker_pid == session.proxy_worker_pid
    assert _session.proxy_transport == session.proxy_transport
    assert _session.realm == session.realm
    assert _session.authid == session.authid
    assert _session.authrole == session.authrole
    assert _session.authmethod == session.authmethod
    assert _session.authprovider == session.authprovider
    assert _session.authextra == session.authextra


def test_session_roundtrip_perf(session, builder):
    obj = session.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"session": 0}

    def loop():
        _session = Session.cast(data)
        if True:
            assert _session.arealm_oid == session.arealm_oid
            assert _session.oid == session.oid
            assert _session.session == session.session
            assert _session.joined_at == session.joined_at
            assert _session.left_at == session.left_at
            assert _session.node_oid == session.node_oid
            assert _session.node_authid == session.node_authid
            assert _session.worker_name == session.worker_name
            assert _session.worker_pid == session.worker_pid
            assert _session.transport == session.transport
            assert _session.proxy_node_oid == session.proxy_node_oid
            assert _session.proxy_node_authid == session.proxy_node_authid
            assert _session.proxy_worker_name == session.proxy_worker_name
            assert _session.proxy_worker_pid == session.proxy_worker_pid
            assert _session.proxy_transport == session.proxy_transport
            assert _session.realm == session.realm
            assert _session.authid == session.authid
            assert _session.authrole == session.authrole
            assert _session.authmethod == session.authmethod
            assert _session.authprovider == session.authprovider
            assert _session.authextra == session.authextra

            scratch["session"] += session.session

    N = 5
    M = 100000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    assert scratch["session"] > 0



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint

from zlmdb import MapOidTimestampFlatBuffers, table

from cfxdb.gen.realmstore import Event as EventGen


class Event(object):
    """
    Persisted event database object.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        self._timestamp = None
        self._subscription = None
        self._publication = None
        self._receiver = None
        self._retained = None
        self._acknowledged_delivery = None

    def marshal(self):
        obj = {
            "timestamp": self.timestamp,
            "subscription": self.subscription,
            "publication": self.publication,
            "receiver": self.receiver,
            "retained": self.retained,
            "acknowledged_delivery": self.acknowledged_delivery,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def timestamp(self):
        """
        Timestamp when the event was sent to the receiver. Epoch time in ns.

        :returns: Epoc time in ns.
        :rtype: int
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = self._from_fbs.Timestamp()
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        assert type(value) == int
        self._timestamp = value

    @property
    def subscription(self):
        """
        The subscription ID this event is dispatched under.

        :returns: The subscription ID.
        :rtype: int
        """
        if self._subscription is None and self._from_fbs:
            self._subscription = self._from_fbs.Subscription()
        return self._subscription

    @subscription.setter
    def subscription(self, value):
        assert type(value) == int
        self._subscription = value

    @property
    def publication(self):
        """
        The publication ID of the dispatched event.

        :returns: The publication ID.
        :rtype: int
        """
        if self._publication is None and self._from_fbs:
            self._publication = self._from_fbs.Publication()
        return self._publication

    @publication.setter
    def publication(self, value):
        assert type(value) == int
        self._publication = value

    @property
    def receiver(self):
        """
        The WAMP session ID of the receiver.

        :returns: The receiver ID.
        :rtype: int
        """
        if self._receiver is None and self._from_fbs:
            self._receiver = self._from_fbs.Receiver()
        return self._receiver

    @receiver.setter
    def receiver(self, value):
        assert type(value) == int
        self._receiver = value

    @property
    def retained(self):
        """
        Whether the message was retained by the broker on the topic, rather than just published.

        :returns: retained flag
        :rtype: bool
        """
        if self._retained is None and self._from_fbs:
            self._retained = self._from_fbs.Retained()
        return self._retained

    @retained.setter
    def retained(self, value):
        assert type(value) == bool
        self._retained = value

    @property
    def acknowledged_delivery(self):
        """
        Whether this Event was to be acknowledged by the receiver.

        :returns: acknowledged delivery flag
        :rtype: bool
        """
        if self._acknowledged_delivery is None and self._from_fbs:
            self._acknowledged_delivery = self._from_fbs.AcknowledgedDelivery()
        return self._acknowledged_delivery

    @acknowledged_delivery.setter
    def acknowledged_delivery(self, value):
        assert type(value) == bool
        self._acknowledged_delivery = value

    @staticmethod
    def cast(buf):
        return Event(EventGen.Event.GetRootAsEvent(buf, 0))

    def build(self, builder):
        # now start and build a new object ..
        EventGen.EventStart(builder)

        EventGen.EventAddTimestamp(builder, self.timestamp)
        EventGen.EventAddSubscription(builder, self.subscription)
        EventGen.EventAddPublication(builder, self.publication)
        EventGen.EventAddReceiver(builder, self.receiver)

        if self.retained is not None:
            EventGen.EventAddRetained(builder, self.retained)
        if self.acknowledged_delivery is not None:
            EventGen.EventAddAcknowledgedDelivery(builder, self.acknowledged_delivery)

        # finish the object.
        final = EventGen.EventEnd(builder)

        return final


@table("40a9df31-6065-496f-809f-027a1879654c", build=Event.build, cast=Event.cast)
class Events(MapOidTimestampFlatBuffers):
    """
    Persisted events archive.


    Map :class:`zlmdb.MapOid3FlatBuffers` from ``(subscription, time_ns)`` to :class:`cfxdb.eventstore.Event`.
    """



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: realmstore

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# This table stores WAMP publications with configurable amount of details, optionally including application payload.
class Publication(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Publication()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsPublication(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Publication
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Timestamp when the publication was accepted by the broker. Epoch time in ns.
    # Publication
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # WAMP publication ID that was assigned by the broker.
    # Publication
    def Publication(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # WAMP session ID of the publisher.
    # Publication
    def Publisher(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # The WAMP or application URI of the PubSub topic the event was published to.
    # Publication
    def Topic(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Positional values for application-defined event payload.
    # Publication
    def Args(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Publication
    def ArgsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Publication
    def ArgsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publication
    def ArgsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Keyword values for application-defined event payload.
    # Publication
    def Kwargs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Publication
    def KwargsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Publication
    def KwargsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publication
    def KwargsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Alternative, transparent payload. If given, ``args`` and ``kwargs`` must be left unset.
    # Publication
    def Payload(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Publication
    def PayloadAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Publication
    def PayloadLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publication
    def PayloadIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # If ``True``, the broker was asked to acknowledge the publication with a success or error response.
    # Publication
    def Acknowledge(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # If ``True``, the broker was requested to retain this event.
    # Publication
    def Retain(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # If ``True``, the broker was asked to exclude the publisher from receiving the event.
    # Publication
    def ExcludeMe(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # List of WAMP session IDs to exclude from receiving this event.
    # Publication
    def Exclude(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8)
            )
        return 0

    # Publication
    def ExcludeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # Publication
    def ExcludeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publication
    def ExcludeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        return o == 0

    # List of WAMP authids to exclude from receiving this event.
    # Publication
    def ExcludeAuthid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Publication
    def ExcludeAuthidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publication
    def ExcludeAuthidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        return o == 0

    # List of WAMP authroles to exclude from receiving this event.
    # Publication
    def ExcludeAuthrole(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Publication
    def ExcludeAuthroleLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publication
    def ExcludeAuthroleIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        return o == 0

    # List of WAMP session IDs eligible to receive this event.
    # Publication
    def Eligible(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8)
            )
        return 0

    # Publication
    def EligibleAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # Publication
    def EligibleLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publication
    def EligibleIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        return o == 0

    # List of WAMP authids eligible to receive this event.
    # Publication
    def EligibleAuthid(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Publication
    def EligibleAuthidLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publication
    def EligibleAuthidIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        return o == 0

    # List of WAMP authroles eligible to receive this event.
    # Publication
    def EligibleAuthrole(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Publication
    def EligibleAuthroleLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publication
    def EligibleAuthroleIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        return o == 0

    # When using payload transparency, the encoding algorithm that was used to encode the payload.
    # Publication
    def EncAlgo(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # When using payload transparency with an encryption algorithm, the payload encryption key.
    # Publication
    def EncKey(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # Publication
    def EncKeyAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Publication
    def EncKeyLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Publication
    def EncKeyIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        return o == 0

    # When using payload transparency, the payload object serializer that was used encoding the payload.
    # Publication
    def EncSerializer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(40))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0


def PublicationStart(builder):
    builder.StartObject(19)


def Start(builder):
    PublicationStart(builder)


def PublicationAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(0, timestamp, 0)


def AddTimestamp(builder, timestamp):
    PublicationAddTimestamp(builder, timestamp)


def PublicationAddPublication(builder, publication):
    builder.PrependUint64Slot(1, publication, 0)


def AddPublication(builder, publication):
    PublicationAddPublication(builder, publication)


def PublicationAddPublisher(builder, publisher):
    builder.PrependUint64Slot(2, publisher, 0)


def AddPublisher(builder, publisher):
    PublicationAddPublisher(builder, publisher)


def PublicationAddTopic(builder, topic):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(topic), 0)


def AddTopic(builder, topic):
    PublicationAddTopic(builder, topic)


def PublicationAddArgs(builder, args):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(args), 0)


def AddArgs(builder, args):
    PublicationAddArgs(builder, args)


def PublicationStartArgsVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartArgsVector(builder, numElems: int) -> int:
    return PublicationStartArgsVector(builder, numElems)


def PublicationAddKwargs(builder, kwargs):
    builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(kwargs), 0)


def AddKwargs(builder, kwargs):
    PublicationAddKwargs(builder, kwargs)


def PublicationStartKwargsVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartKwargsVector(builder, numElems: int) -> int:
    return PublicationStartKwargsVector(builder, numElems)


def PublicationAddPayload(builder, payload):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(payload), 0)


def AddPayload(builder, payload):
    PublicationAddPayload(builder, payload)


def PublicationStartPayloadVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPayloadVector(builder, numElems: int) -> int:
    return PublicationStartPayloadVector(builder, numElems)


def PublicationAddAcknowledge(builder, acknowledge):
    builder.PrependBoolSlot(7, acknowledge, 0)


def AddAcknowledge(builder, acknowledge):
    PublicationAddAcknowledge(builder, acknowledge)


def PublicationAddRetain(builder, retain):
    builder.PrependBoolSlot(8, retain, 0)


def AddRetain(builder, retain):
    PublicationAddRetain(builder, retain)


def PublicationAddExcludeMe(builder, excludeMe):
    builder.PrependBoolSlot(9, excludeMe, 0)


def AddExcludeMe(builder, excludeMe):
    PublicationAddExcludeMe(builder, excludeMe)


def PublicationAddExclude(builder, exclude):
    builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(exclude), 0)


def AddExclude(builder, exclude):
    PublicationAddExclude(builder, exclude)


def PublicationStartExcludeVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)


def StartExcludeVector(builder, numElems: int) -> int:
    return PublicationStartExcludeVector(builder, numElems)


def PublicationAddExcludeAuthid(builder, excludeAuthid):
    builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(excludeAuthid), 0)


def AddExcludeAuthid(builder, excludeAuthid):
    PublicationAddExcludeAuthid(builder, excludeAuthid)


def PublicationStartExcludeAuthidVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartExcludeAuthidVector(builder, numElems: int) -> int:
    return PublicationStartExcludeAuthidVector(builder, numElems)


def PublicationAddExcludeAuthrole(builder, excludeAuthrole):
    builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(excludeAuthrole), 0)


def AddExcludeAuthrole(builder, excludeAuthrole):
    PublicationAddExcludeAuthrole(builder, excludeAuthrole)


def PublicationStartExcludeAuthroleVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartExcludeAuthroleVector(builder, numElems: int) -> int:
    return PublicationStartExcludeAuthroleVector(builder, numElems)


def PublicationAddEligible(builder, eligible):
    builder.PrependUOffsetTRelativeSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(eligible), 0)


def AddEligible(builder, eligible):
    PublicationAddEligible(builder, eligible)


def PublicationStartEligibleVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)


def StartEligibleVector(builder, numElems: int) -> int:
    return PublicationStartEligibleVector(builder, numElems)


def PublicationAddEligibleAuthid(builder, eligibleAuthid):
    builder.PrependUOffsetTRelativeSlot(14, flatbuffers.number_types.UOffsetTFlags.py_type(eligibleAuthid), 0)


def AddEligibleAuthid(builder, eligibleAuthid):
    PublicationAddEligibleAuthid(builder, eligibleAuthid)


def PublicationStartEligibleAuthidVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartEligibleAuthidVector(builder, numElems: int) -> int:
    return PublicationStartEligibleAuthidVector(builder, numElems)


def PublicationAddEligibleAuthrole(builder, eligibleAuthrole):
    builder.PrependUOffsetTRelativeSlot(15, flatbuffers.number_types.UOffsetTFlags.py_type(eligibleAuthrole), 0)


def AddEligibleAuthrole(builder, eligibleAuthrole):
    PublicationAddEligibleAuthrole(builder, eligibleAuthrole)


def PublicationStartEligibleAuthroleVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartEligibleAuthroleVector(builder, numElems: int) -> int:
    return PublicationStartEligibleAuthroleVector(builder, numElems)


def PublicationAddEncAlgo(builder, encAlgo):
    builder.PrependUint8Slot(16, encAlgo, 0)


def AddEncAlgo(builder, encAlgo):
    PublicationAddEncAlgo(builder, encAlgo)


def PublicationAddEncKey(builder, encKey):
    builder.PrependUOffsetTRelativeSlot(17, flatbuffers.number_types.UOffsetTFlags.py_type(encKey), 0)


def AddEncKey(builder, encKey):
    PublicationAddEncKey(builder, encKey)


def PublicationStartEncKeyVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartEncKeyVector(builder, numElems: int) -> int:
    return PublicationStartEncKeyVector(builder, numElems)


def PublicationAddEncSerializer(builder, encSerializer):
    builder.PrependUint8Slot(18, encSerializer, 0)


def AddEncSerializer(builder, encSerializer):
    PublicationAddEncSerializer(builder, encSerializer)


def PublicationEnd(builder):
    return builder.EndObject()


def End(builder):
    return PublicationEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.gen.xbrmm.ChannelState import ChannelState
from cfxdb.gen.xbrmm.ChannelType import ChannelType
from cfxdb.gen.xbrmm.TransactionState import TransactionState
from cfxdb.xbrmm.channel import (
    Channel,
    ChannelBalance,
    IndexPayingChannelByDelegate,
    IndexPayingChannelByRecipient,
    IndexPaymentChannelByDelegate,
    PayingChannelBalances,
    PayingChannels,
    PaymentChannelBalances,
    PaymentChannels,
)
from cfxdb.xbrmm.channel import Channel as PayingChannel
from cfxdb.xbrmm.channel import Channel as PaymentChannel
from cfxdb.xbrmm.channel import ChannelBalance as PayingChannelBalance
from cfxdb.xbrmm.channel import ChannelBalance as PaymentChannelBalance
from cfxdb.xbrmm.ipfs_file import IPFSFile, IPFSFiles
from cfxdb.xbrmm.offer import IndexOfferByKey, Offer, Offers
from cfxdb.xbrmm.schema import Schema
from cfxdb.xbrmm.transaction import Transaction, Transactions
from cfxdb.xbrmm.userkey import IndexUserKeyByMember, UserKey, UserKeys

__all__ = (
    # database schema
    "Schema",
    # enum types
    "ChannelType",
    "ChannelState",
    "TransactionState",
    # table/index types
    "Channel",
    "PaymentChannel",
    "PaymentChannels",
    "IndexPaymentChannelByDelegate",
    "ChannelBalance",
    "PaymentChannelBalance",
    "PaymentChannelBalances",
    "PayingChannel",
    "PayingChannels",
    "IndexPayingChannelByDelegate",
    "IndexPayingChannelByRecipient",
    "PayingChannelBalance",
    "PayingChannelBalances",
    "IPFSFile",
    "IPFSFiles",
    "Offer",
    "Offers",
    "IndexOfferByKey",
    "Transaction",
    "Transactions",
    "UserKey",
    "UserKeys",
    "IndexUserKeyByMember",
)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# This table stores XBRToken.Transfer events.
class TokenTransfer(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TokenTransfer()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTokenTransfer(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # TokenTransfer
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Primary key: Transaction hash.
    # TokenTransfer
    def TxHash(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenTransfer
    def TxHashAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenTransfer
    def TxHashLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenTransfer
    def TxHashIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Block hash.
    # TokenTransfer
    def BlockHash(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenTransfer
    def BlockHashAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenTransfer
    def BlockHashLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenTransfer
    def BlockHashIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # XBR token sending address.
    # TokenTransfer
    def FromAddress(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenTransfer
    def FromAddressAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenTransfer
    def FromAddressLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenTransfer
    def FromAddressIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # XBR token receiving address.
    # TokenTransfer
    def ToAddress(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenTransfer
    def ToAddressAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenTransfer
    def ToAddressLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenTransfer
    def ToAddressIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # XBR token transferred.
    # TokenTransfer
    def Value(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenTransfer
    def ValueAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenTransfer
    def ValueLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenTransfer
    def ValueIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0


def TokenTransferStart(builder):
    builder.StartObject(5)


def Start(builder):
    TokenTransferStart(builder)


def TokenTransferAddTxHash(builder, txHash):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(txHash), 0)


def AddTxHash(builder, txHash):
    TokenTransferAddTxHash(builder, txHash)


def TokenTransferStartTxHashVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTxHashVector(builder, numElems: int) -> int:
    return TokenTransferStartTxHashVector(builder, numElems)


def TokenTransferAddBlockHash(builder, blockHash):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(blockHash), 0)


def AddBlockHash(builder, blockHash):
    TokenTransferAddBlockHash(builder, blockHash)


def TokenTransferStartBlockHashVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartBlockHashVector(builder, numElems: int) -> int:
    return TokenTransferStartBlockHashVector(builder, numElems)


def TokenTransferAddFromAddress(builder, fromAddress):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(fromAddress), 0)


def AddFromAddress(builder, fromAddress):
    TokenTransferAddFromAddress(builder, fromAddress)


def TokenTransferStartFromAddressVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartFromAddressVector(builder, numElems: int) -> int:
    return TokenTransferStartFromAddressVector(builder, numElems)


def TokenTransferAddToAddress(builder, toAddress):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(toAddress), 0)


def AddToAddress(builder, toAddress):
    TokenTransferAddToAddress(builder, toAddress)


def TokenTransferStartToAddressVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartToAddressVector(builder, numElems: int) -> int:
    return TokenTransferStartToAddressVector(builder, numElems)


def TokenTransferAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0)


def AddValue(builder, value):
    TokenTransferAddValue(builder, value)


def TokenTransferStartValueVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartValueVector(builder, numElems: int) -> int:
    return TokenTransferStartValueVector(builder, numElems)


def TokenTransferEnd(builder):
    return builder.EndObject()


def End(builder):
    return TokenTransferEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import binascii
import os
import uuid
from datetime import datetime

import cbor2
import pytest
import txaio
from autobahn import util
from zlmdb import flatbuffers

from cfxdb.user import ActivationToken, ActivationTokenFbs

txaio.use_twisted()


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def fill_token(token):
    token.oid = uuid.uuid4()
    token.atype = 1
    token.status = 1
    token.created = datetime.utcnow()
    token.completed = None
    token.code = util.generate_activation_code()
    token.email = "homer.simpson@example.com"
    token.pubkey = binascii.b2a_hex(os.urandom(32)).decode()


@pytest.fixture(scope="function")
def token_fbs():
    _token = ActivationTokenFbs()
    fill_token(_token)
    return _token


@pytest.fixture(scope="function")
def token_cbor():
    _token = ActivationToken()
    fill_token(_token)
    return _token


def test_token_fbs_roundtrip(token_fbs, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = token_fbs.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 224

    # create python object from bytes (flatbuffes)
    _token = ActivationTokenFbs.cast(data)

    # assert _token == token_fbs

    assert _token.oid == token_fbs.oid
    assert _token.atype == token_fbs.atype
    assert _token.status == token_fbs.status
    assert _token.completed == token_fbs.completed
    assert _token.code == token_fbs.code
    assert _token.email == token_fbs.email
    assert _token.pubkey == token_fbs.pubkey


def test_token_cbor_roundtrip(token_cbor):
    # serialize to bytes (cbor) from python object
    obj = token_cbor.marshal()
    data = cbor2.dumps(obj)
    assert len(data) == 212

    # create python object from bytes (cbor)
    _obj = cbor2.loads(data)
    _token = ActivationToken.parse(_obj)

    # assert _token == token_cbor

    assert _token.oid == token_cbor.oid
    assert _token.atype == token_cbor.atype
    assert _token.status == token_cbor.status
    assert _token.completed == token_cbor.completed
    assert _token.code == token_cbor.code
    assert _token.email == token_cbor.email
    assert _token.pubkey == token_cbor.pubkey


def test_token_copy_cbor_to_fbs(token_cbor):
    token_fbs = ActivationTokenFbs()
    token_fbs.copy(token_cbor)

    # assert token_fbs == token_cbor

    assert token_fbs.oid == token_cbor.oid
    assert token_fbs.atype == token_cbor.atype
    assert token_fbs.status == token_cbor.status
    assert token_fbs.completed == token_cbor.completed
    assert token_fbs.code == token_cbor.code
    assert token_fbs.email == token_cbor.email
    assert token_fbs.pubkey == token_cbor.pubkey


def test_token_copy_fbs_to_cbor(token_fbs):
    token_cbor = ActivationToken()
    token_cbor.copy(token_fbs)

    # assert token_cbor == token_fbs

    assert token_cbor.oid == token_fbs.oid
    assert token_cbor.atype == token_fbs.atype
    assert token_cbor.status == token_fbs.status
    assert token_cbor.completed == token_fbs.completed
    assert token_cbor.code == token_fbs.code
    assert token_cbor.email == token_fbs.email
    assert token_cbor.pubkey == token_fbs.pubkey



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit
import uuid

import numpy as np
import pytest
import txaio
import zlmdb
from txaio import time_ns
from zlmdb import flatbuffers

from cfxdb.usage import MasterNodeUsage

zlmdb.TABLES_BY_UUID = {}
txaio.use_twisted()


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


#
# Usage
#


def fill_usage(usage):
    usage.timestamp = np.datetime64(time_ns(), "ns")
    usage.mrealm_id = uuid.uuid4()
    usage.timestamp_from = np.datetime64(usage.timestamp - np.timedelta64(10, "m"), "ns")
    usage.pubkey = os.urandom(32)

    usage.client_ip_version = random.choice([4, 6])
    if usage.client_ip_version == 4:
        usage.client_ip_address = os.urandom(4)
    else:
        usage.client_ip_address = os.urandom(16)
    usage.client_ip_port = random.randint(1, 2**16 - 1)

    usage.seq = random.randint(0, 1000000)
    usage.sent = np.datetime64(time_ns() - random.randint(0, 10**10), "ns")
    usage.processed = np.datetime64(time_ns() + random.randint(0, 10**10), "ns")
    usage.status = random.randint(0, 3)
    usage.status_message = "hello world {}".format(uuid.uuid4())
    usage.metering_id = uuid.uuid4()

    usage.count = random.randint(0, 100000)
    usage.total = random.randint(0, 100000)
    usage.nodes = random.randint(0, 100000)

    usage.controllers = random.randint(0, 100000)
    usage.hostmonitors = random.randint(0, 100000)
    usage.routers = random.randint(0, 100000)
    usage.containers = random.randint(0, 100000)
    usage.guests = random.randint(0, 100000)
    usage.proxies = random.randint(0, 100000)
    usage.marketmakers = random.randint(0, 100000)

    usage.sessions = random.randint(0, 100000)

    usage.msgs_call = random.randint(0, 100000)
    usage.msgs_yield = random.randint(0, 100000)
    usage.msgs_invocation = random.randint(0, 100000)
    usage.msgs_result = random.randint(0, 100000)
    usage.msgs_error = random.randint(0, 100000)
    usage.msgs_publish = random.randint(0, 100000)
    usage.msgs_published = random.randint(0, 100000)
    usage.msgs_event = random.randint(0, 100000)
    usage.msgs_register = random.randint(0, 100000)
    usage.msgs_registered = random.randint(0, 100000)
    usage.msgs_subscribe = random.randint(0, 100000)
    usage.msgs_subscribed = random.randint(0, 100000)


@pytest.fixture(scope="function")
def usage():
    _usage = MasterNodeUsage()
    fill_usage(_usage)
    return _usage


def test_usage_roundtrip(usage, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = usage.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) in [520, 512, 504]

    # create python object from bytes (flatbuffes)
    _usage = MasterNodeUsage.cast(data)

    assert _usage.timestamp == _usage.timestamp
    assert _usage.mrealm_id == _usage.mrealm_id
    assert _usage.timestamp_from == _usage.timestamp_from
    assert _usage.pubkey == _usage.pubkey
    assert _usage.client_ip_address == _usage.client_ip_address
    assert _usage.client_ip_version == _usage.client_ip_version
    assert _usage.client_ip_port == _usage.client_ip_port
    assert _usage.seq == _usage.seq
    assert _usage.sent == _usage.sent
    assert _usage.processed == _usage.processed
    assert _usage.status == _usage.status
    assert _usage.status_message == _usage.status_message
    assert _usage.metering_id == _usage.metering_id

    assert _usage.count == _usage.count
    assert _usage.total == _usage.total
    assert _usage.nodes == _usage.nodes
    assert _usage.controllers == _usage.controllers
    assert _usage.hostmonitors == _usage.hostmonitors
    assert _usage.routers == _usage.routers
    assert _usage.containers == _usage.containers
    assert _usage.guests == _usage.guests
    assert _usage.proxies == _usage.proxies
    assert _usage.marketmakers == _usage.marketmakers
    assert _usage.sessions == _usage.sessions
    assert _usage.msgs_call == _usage.msgs_call
    assert _usage.msgs_yield == _usage.msgs_yield
    assert _usage.msgs_invocation == _usage.msgs_invocation
    assert _usage.msgs_result == _usage.msgs_result
    assert _usage.msgs_error == _usage.msgs_error
    assert _usage.msgs_publish == _usage.msgs_publish
    assert _usage.msgs_published == _usage.msgs_published
    assert _usage.msgs_event == _usage.msgs_event
    assert _usage.msgs_register == _usage.msgs_register
    assert _usage.msgs_registered == _usage.msgs_registered
    assert _usage.msgs_subscribe == _usage.msgs_subscribe
    assert _usage.msgs_subscribed == _usage.msgs_subscribed


def test_usage_roundtrip_perf(usage, builder):
    obj = usage.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {}

    def loop():
        _usage = MasterNodeUsage.cast(data)
        if True:
            assert _usage.timestamp == _usage.timestamp
            assert _usage.mrealm_id == _usage.mrealm_id
            assert _usage.timestamp_from == _usage.timestamp_from
            assert _usage.pubkey == _usage.pubkey
            assert _usage.client_ip_address == _usage.client_ip_address
            assert _usage.client_ip_version == _usage.client_ip_version
            assert _usage.client_ip_port == _usage.client_ip_port
            assert _usage.seq == _usage.seq
            assert _usage.sent == _usage.sent
            assert _usage.processed == _usage.processed
            assert _usage.status == _usage.status
            assert _usage.status_message == _usage.status_message
            assert _usage.metering_id == _usage.metering_id

            assert _usage.count == _usage.count
            assert _usage.total == _usage.total
            assert _usage.nodes == _usage.nodes
            assert _usage.controllers == _usage.controllers
            assert _usage.hostmonitors == _usage.hostmonitors
            assert _usage.routers == _usage.routers
            assert _usage.containers == _usage.containers
            assert _usage.guests == _usage.guests
            assert _usage.proxies == _usage.proxies
            assert _usage.marketmakers == _usage.marketmakers
            assert _usage.sessions == _usage.sessions
            assert _usage.msgs_call == _usage.msgs_call
            assert _usage.msgs_yield == _usage.msgs_yield
            assert _usage.msgs_invocation == _usage.msgs_invocation
            assert _usage.msgs_result == _usage.msgs_result
            assert _usage.msgs_error == _usage.msgs_error
            assert _usage.msgs_publish == _usage.msgs_publish
            assert _usage.msgs_published == _usage.msgs_published
            assert _usage.msgs_event == _usage.msgs_event
            assert _usage.msgs_register == _usage.msgs_register
            assert _usage.msgs_registered == _usage.msgs_registered
            assert _usage.msgs_subscribe == _usage.msgs_subscribe
            assert _usage.msgs_subscribed == _usage.msgs_subscribed

            # copy, because of "cannot hash writable memoryview object"
            pubkey = bytes(_usage.pubkey)

            if pubkey not in scratch:
                scratch[pubkey] = 0
            scratch[pubkey] += 1

    N = 5
    M = 50000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    assert len(scratch) > 0



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit
import uuid

import txaio

txaio.use_twisted()  # noqa

import numpy as np
import pytest
from txaio import time_ns
from zlmdb import flatbuffers

from cfxdb.tests._util import _gen_ipfs_hash
from cfxdb.xbr import Member


def fill_member(member):
    member.address = os.urandom(20)
    member.account_oid = uuid.uuid4()
    member.timestamp = np.datetime64(time_ns(), "ns")
    member.registered = random.randint(0, 2**256 - 1)
    member.eula = _gen_ipfs_hash()
    member.profile = _gen_ipfs_hash()
    member.level = random.randint(1, 5)
    member.tid = os.urandom(32)
    member.signature = os.urandom(65)


@pytest.fixture(scope="function")
def member():
    _member = Member()
    fill_member(_member)
    return _member


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def test_member_roundtrip(member, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = member.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 384

    # create python object from bytes (flatbuffes)
    _member = Member.cast(data)

    assert _member.address == member.address
    assert _member.account_oid == member.account_oid
    assert _member.timestamp == member.timestamp
    assert _member.registered == member.registered
    assert _member.eula == member.eula
    assert _member.profile == member.profile
    assert _member.level == member.level
    assert _member.tid == member.tid
    assert _member.signature == member.signature


def test_member_roundtrip_perf(member, builder):
    obj = member.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _member = Member.cast(data)
        if True:
            assert _member.address == member.address
            assert _member.account_oid == member.account_oid
            assert _member.timestamp == member.timestamp
            assert _member.registered == member.registered
            assert _member.eula == member.eula
            assert _member.profile == member.profile
            assert _member.level == member.level
            assert _member.tid == member.tid
            assert _member.signature == member.signature

            scratch["value"] += _member.level

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from uuid import UUID

import cbor2
import numpy as np
from zlmdb import MapUuidUuidStringFlatBuffers, flatbuffers, table

from cfxdb.gen.meta import Attribute as AttributeGen


class _AttributeGen(AttributeGen.Attribute):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsAttribute(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _AttributeGen()
        x.Init(buf, n + offset)
        return x

    def TableOidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ObjectOidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ValueAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Attribute(object):
    """
    Generic meta-data attributes that can be stored on objects in tables.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # [uint8] (required, uuid)
        self._table_oid = None

        # [uint8] (required, uuid)
        self._object_oid = None

        # string (required)
        self._attribute = None

        # uint64 (timestamp)
        self._modified = None

        # [uint8] (cbor)
        self._value = None

    def marshal(self) -> dict:
        obj = {
            "table_oid": self.table_oid.bytes if self.object_oid else None,
            "object_oid": self.object_oid.bytes if self.object_oid else None,
            "attribute": self.attribute,
            "modified": int(self.modified) if self.modified else None,
            "value": bytes(self._value) if self._value else None,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def table_oid(self) -> UUID:
        """
        Table of the object holding the attribute.
        """
        if self._table_oid is None and self._from_fbs:
            if self._from_fbs.TableOidLength():
                _table_oid = self._from_fbs.TableOidAsBytes()
                self._table_oid = UUID(bytes=bytes(_table_oid))
        return self._table_oid

    @table_oid.setter
    def table_oid(self, value: UUID):
        assert value is None or isinstance(value, UUID)
        self._table_oid = value

    @property
    def object_oid(self) -> UUID:
        """
        Object (within the table) holding the attribute.
        """
        if self._object_oid is None and self._from_fbs:
            if self._from_fbs.ObjectOidLength():
                _object_oid = self._from_fbs.ObjectOidAsBytes()
                self._object_oid = UUID(bytes=bytes(_object_oid))
        return self._object_oid

    @object_oid.setter
    def object_oid(self, value: UUID):
        assert value is None or isinstance(value, UUID)
        self._object_oid = value

    @property
    def attribute(self) -> str:
        """
        Attribute name (part of the key).
        """
        if self._attribute is None and self._from_fbs:
            attribute = self._from_fbs.Attribute()
            if attribute:
                self._attribute = attribute.decode("utf8")
        return self._attribute

    @attribute.setter
    def attribute(self, value: str):
        assert value is None or type(value) == str
        self._attribute = value

    @property
    def modified(self) -> np.datetime64:
        """
        Timestamp when the attribute was last modified (or first created).
        """
        if self._modified is None and self._from_fbs:
            self._modified = np.datetime64(self._from_fbs.Modified(), "ns")
        return self._modified

    @modified.setter
    def modified(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._modified = value

    @property
    def value(self):
        """
        Arbitrary attribute value, stored CBOR-serialized.
        """
        if self._value is None and self._from_fbs:
            if self._from_fbs.ValueLength():
                _value = self._from_fbs.ValueAsBytes()
                if _value:
                    self._value = cbor2.loads(_value)
        return self._value

    @value.setter
    def value(self, new_value):
        self._value = new_value

    @staticmethod
    def cast(buf):
        return Attribute(_AttributeGen.GetRootAsAttribute(buf, 0))

    def build(self, builder):
        table_oid = self.table_oid.bytes if self.table_oid else None
        if table_oid:
            table_oid = builder.CreateString(table_oid)

        object_oid = self.object_oid.bytes if self.object_oid else None
        if object_oid:
            object_oid = builder.CreateString(object_oid)

        attribute = self.attribute
        if attribute:
            attribute = builder.CreateString(attribute)

        value = self.value
        if value:
            value = builder.CreateString(cbor2.dumps(value))

        AttributeGen.AttributeStart(builder)

        if table_oid:
            AttributeGen.AttributeAddTableOid(builder, table_oid)

        if object_oid:
            AttributeGen.AttributeAddObjectOid(builder, object_oid)

        if attribute:
            AttributeGen.AttributeAddAttribute(builder, attribute)

        if self.modified:
            AttributeGen.AttributeAddModified(builder, int(self.modified))

        if value:
            AttributeGen.AttributeAddValue(builder, value)

        final = AttributeGen.AttributeEnd(builder)

        return final


@table("42b1ca1f-f135-4761-8d10-f96a43612178", build=Attribute.build, cast=Attribute.cast)
class Attributes(MapUuidUuidStringFlatBuffers):
    """
    Generic meta-data attributes that can be stored on objects in tables. Primary key of this table is ``(table_oid, object_oid, attribute)``.

    Map :class:`zlmdb.MapUuidUuidStringFlatBuffers` from ``(table_oid, object_oid, attribute)`` to :class:`cfxdb.meta.Attribute`
    """



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import List, Optional
from uuid import UUID

import numpy as np

from cfxdb.common import ConfigurationElement


class Role(ConfigurationElement):
    """
    Roles created for use with application-level authorization and permissions
    in application realms.
    """

    def __init__(
        self,
        oid: Optional[UUID] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        name: Optional[str] = None,
        created: Optional[np.datetime64] = None,
        owner: Optional[UUID] = None,
        _unknown=None,
    ):
        """

        :param oid: Object ID of role

        :param label: Optional user label of role

        :param description: Optional user description of role

        :param tags: Optional list of user tags on role

        :param name: Name of role

        :param created: Timestamp when the role was created

        :param owner: Owning user (object ID)
        """
        ConfigurationElement.__init__(self, oid=oid, label=label, description=description, tags=tags)
        self.name = name
        self.created = created
        self.owner = owner

        # private member with unknown/untouched data passing through
        self._unknown = _unknown

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if not ConfigurationElement.__eq__(self, other):
            return False
        if other.name != self.name:
            return False
        if other.created != self.created:
            return False
        if other.owner != self.owner:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def copy(self, other, overwrite=False):
        """
        Copy over other object.

        :param other: Other role to copy data from.
        :type other: instance of :class:`ManagementRealm`
        :return:
        """
        ConfigurationElement.copy(self, other, overwrite=overwrite)

        if (not self.name and other.name) or overwrite:
            self.name = other.name
        if (not self.created and other.created) or overwrite:
            self.created = other.created
        if (not self.owner and other.owner) or overwrite:
            self.owner = other.owner

        # _unknown is not copied!

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = ConfigurationElement.marshal(self)

        obj.update(
            {
                "oid": str(self.oid) if self.oid else None,
                "name": self.name,
                "created": int(self.created) if self.created else None,
                "owner": str(self.owner) if self.owner else None,
            }
        )

        if self._unknown:
            # pass through all attributes unknown
            obj.update(self._unknown)

        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`ManagementRealm`
        """
        assert type(data) == dict

        obj = ConfigurationElement.parse(data)
        data = obj._unknown

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["oid", "name", "owner", "created"]:
                _unknown[k] = data[k]

        name = data.get("name", None)
        assert name is None or type(name) == str

        owner = data.get("owner", None)
        assert owner is None or type(owner) == str
        if owner:
            owner = UUID(owner)

        created = data.get("created", None)
        assert created is None or type(created) == int
        if created:
            created = np.datetime64(created, "ns")

        obj = Role(
            oid=obj.oid,
            label=obj.label,
            description=obj.description,
            tags=obj.tags,
            name=name,
            owner=owner,
            created=created,
            _unknown=_unknown,
        )

        return obj



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit
import uuid

import txaio

txaio.use_twisted()  # noqa

import numpy as np
import pytest
from autobahn import util
from txaio import time_ns
from zlmdb import flatbuffers

from cfxdb.xbrmm import Offer


def fill_offer(offer):
    now = time_ns()
    offer.timestamp = np.datetime64(now, "ns")
    offer.offer = uuid.uuid4()
    offer.seller = os.urandom(20)
    offer.seller_session_id = random.randint(0, 2**53)
    offer.seller_authid = util.generate_token(5, 4)
    offer.key = uuid.uuid4()
    offer.api = uuid.uuid4()
    offer.uri = "com.example.something.add2"
    offer.valid_from = np.datetime64(now, "ns")
    offer.signature = os.urandom(64)
    offer.price = random.randint(0, 2**256 - 1)
    offer.categories = {
        "xtile": "{:05}".format(random.randint(0, 99999)),
        "ytile": "{:05}".format(random.randint(0, 99999)),
    }
    offer.expires = np.datetime64(now + 60 * 60 * 10**9, "ns")
    offer.copies = 1000
    offer.remaining = random.randint(0, 1000)


def fill_offer_empty(offer):
    offer.timestamp = None
    offer.offer = None
    offer.seller = None
    offer.seller_session_id = None
    offer.seller_authid = None
    offer.key = None
    offer.api = None
    offer.uri = None
    offer.valid_from = None
    offer.signature = None
    offer.price = None
    offer.categories = None
    offer.expires = None
    offer.copies = None
    offer.remaining = None


@pytest.fixture(scope="function")
def offer():
    _offer = Offer()
    fill_offer(_offer)
    return _offer


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def test_offer_roundtrip(offer, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = offer.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 480

    # create python object from bytes (flatbuffes)
    _offer = Offer.cast(data)

    assert _offer.timestamp == offer.timestamp
    assert _offer.offer == offer.offer
    assert _offer.seller == offer.seller
    assert _offer.seller_session_id == offer.seller_session_id
    assert _offer.seller_authid == offer.seller_authid
    assert _offer.key == offer.key
    assert _offer.api == offer.api
    assert _offer.uri == offer.uri
    assert _offer.valid_from == offer.valid_from
    assert _offer.signature == offer.signature
    assert _offer.price == offer.price
    assert _offer.categories == offer.categories
    assert _offer.expires == offer.expires
    assert _offer.copies == offer.copies
    assert _offer.remaining == offer.remaining


def test_offer_empty(builder):
    offer = Offer()
    fill_offer_empty(offer)

    # serialize to bytes (flatbuffers) from python object
    obj = offer.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 12

    # create python object from bytes (flatbuffes)
    _offer = Offer.cast(data)

    unix_zero = np.datetime64(0, "ns")

    assert _offer.timestamp == unix_zero
    assert _offer.offer is None
    assert _offer.seller is None
    assert _offer.seller_session_id == 0
    assert _offer.seller_authid is None
    assert _offer.key is None
    assert _offer.api is None
    assert _offer.uri is None
    assert _offer.valid_from == unix_zero
    assert _offer.signature is None
    assert _offer.price == 0
    assert _offer.categories is None
    assert _offer.expires == unix_zero
    assert _offer.copies == 0
    assert _offer.remaining == 0


def test_offer_roundtrip_perf(offer, builder):
    obj = offer.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _offer = Offer.cast(data)
        if True:
            assert _offer.timestamp == offer.timestamp
            assert _offer.offer == offer.offer
            assert _offer.seller == offer.seller
            assert _offer.seller_session_id == offer.seller_session_id
            assert _offer.seller_authid == offer.seller_authid
            assert _offer.key == offer.key
            assert _offer.api == offer.api
            assert _offer.uri == offer.uri
            assert _offer.valid_from == offer.valid_from
            assert _offer.signature == offer.signature
            assert _offer.price == offer.price
            assert _offer.categories == offer.categories
            assert _offer.expires == offer.expires
            assert _offer.copies == offer.copies
            assert _offer.remaining == offer.remaining

            scratch["value"] += _offer.price

    N = 7
    M = 20000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrmm

# State of a XBR channel.
class ChannelState(object):
    # Unset
    NONE = 0
    # The channel is open on-chain, and now operating off-chain until closed.
    OPEN = 1
    # The off-chain channel is closing (one of the channel participants has requested to closed the channel)
    CLOSING = 2
    # The off-chain channel is closed and settled with no errors. The final channel state is committed on-chain.
    CLOSED = 3
    # The channel is closed with an error. The final channel state is committed on-chain.
    FAILED = 4



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# This table stores XBRToken.Approval events.
class TokenApproval(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TokenApproval()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTokenApproval(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # TokenApproval
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Primary key: Transaction hash.
    # TokenApproval
    def TxHash(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenApproval
    def TxHashAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenApproval
    def TxHashLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenApproval
    def TxHashIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Block hash.
    # TokenApproval
    def BlockHash(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenApproval
    def BlockHashAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenApproval
    def BlockHashLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenApproval
    def BlockHashIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # XBR token sending address.
    # TokenApproval
    def OwnerAddress(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenApproval
    def OwnerAddressAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenApproval
    def OwnerAddressLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenApproval
    def OwnerAddressIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # XBR token receiving address.
    # TokenApproval
    def SpenderAddress(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenApproval
    def SpenderAddressAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenApproval
    def SpenderAddressLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenApproval
    def SpenderAddressIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # XBR token transferred.
    # TokenApproval
    def Value(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenApproval
    def ValueAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenApproval
    def ValueLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenApproval
    def ValueIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0


def TokenApprovalStart(builder):
    builder.StartObject(5)


def Start(builder):
    TokenApprovalStart(builder)


def TokenApprovalAddTxHash(builder, txHash):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(txHash), 0)


def AddTxHash(builder, txHash):
    TokenApprovalAddTxHash(builder, txHash)


def TokenApprovalStartTxHashVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartTxHashVector(builder, numElems: int) -> int:
    return TokenApprovalStartTxHashVector(builder, numElems)


def TokenApprovalAddBlockHash(builder, blockHash):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(blockHash), 0)


def AddBlockHash(builder, blockHash):
    TokenApprovalAddBlockHash(builder, blockHash)


def TokenApprovalStartBlockHashVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartBlockHashVector(builder, numElems: int) -> int:
    return TokenApprovalStartBlockHashVector(builder, numElems)


def TokenApprovalAddOwnerAddress(builder, ownerAddress):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(ownerAddress), 0)


def AddOwnerAddress(builder, ownerAddress):
    TokenApprovalAddOwnerAddress(builder, ownerAddress)


def TokenApprovalStartOwnerAddressVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOwnerAddressVector(builder, numElems: int) -> int:
    return TokenApprovalStartOwnerAddressVector(builder, numElems)


def TokenApprovalAddSpenderAddress(builder, spenderAddress):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(spenderAddress), 0)


def AddSpenderAddress(builder, spenderAddress):
    TokenApprovalAddSpenderAddress(builder, spenderAddress)


def TokenApprovalStartSpenderAddressVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartSpenderAddressVector(builder, numElems: int) -> int:
    return TokenApprovalStartSpenderAddressVector(builder, numElems)


def TokenApprovalAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0)


def AddValue(builder, value):
    TokenApprovalAddValue(builder, value)


def TokenApprovalStartValueVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartValueVector(builder, numElems: int) -> int:
    return TokenApprovalStartValueVector(builder, numElems)


def TokenApprovalEnd(builder):
    return builder.EndObject()


def End(builder):
    return TokenApprovalEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrmm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# Current channel off-chain balance. The sum of ``Balance.remaining`` and ``Balance.inflight`` equals ``Channel.amount``.
class ChannelBalance(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ChannelBalance()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsChannelBalance(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # ChannelBalance
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Amount of ERC20 tokens (of the market coin type) currently remaining in the channel.
    # ChannelBalance
    def Remaining(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # ChannelBalance
    def RemainingAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # ChannelBalance
    def RemainingLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ChannelBalance
    def RemainingIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Amount of ERC20 tokens (of the market coin type) reserved to in-flight transactions.
    # ChannelBalance
    def Inflight(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # ChannelBalance
    def InflightAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # ChannelBalance
    def InflightLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ChannelBalance
    def InflightIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Sequence number of transactions on this balance starting from 0 when the channel is created.
    # ChannelBalance
    def Seq(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0


def ChannelBalanceStart(builder):
    builder.StartObject(3)


def Start(builder):
    ChannelBalanceStart(builder)


def ChannelBalanceAddRemaining(builder, remaining):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(remaining), 0)


def AddRemaining(builder, remaining):
    ChannelBalanceAddRemaining(builder, remaining)


def ChannelBalanceStartRemainingVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartRemainingVector(builder, numElems: int) -> int:
    return ChannelBalanceStartRemainingVector(builder, numElems)


def ChannelBalanceAddInflight(builder, inflight):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(inflight), 0)


def AddInflight(builder, inflight):
    ChannelBalanceAddInflight(builder, inflight)


def ChannelBalanceStartInflightVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartInflightVector(builder, numElems: int) -> int:
    return ChannelBalanceStartInflightVector(builder, numElems)


def ChannelBalanceAddSeq(builder, seq):
    builder.PrependUint32Slot(2, seq, 0)


def AddSeq(builder, seq):
    ChannelBalanceAddSeq(builder, seq)


def ChannelBalanceEnd(builder):
    return builder.EndObject()


def End(builder):
    return ChannelBalanceEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import struct
import uuid

import web3


def unpack_uint256(data):
    assert data is None or type(data) == bytes, "data must by bytes, was {}".format(type(data))
    if data and type(data) == bytes:
        assert len(data) == 32, "data must be bytes[32], but was bytes[{}]".format(len(data))

    if data:
        return web3.Web3.toInt(data)
    else:
        return 0


def pack_uint256(value):
    assert value is None or (type(value) == int and value >= 0 and value < 2**256), (
        "value must be uint256, but was {}".format(value)
    )

    if value:
        data = web3.Web3.toBytes(value)
        return b"\x00" * (32 - len(data)) + data
    else:
        return b"\x00" * 32


class uint256(object):
    def __init__(self, data=None):
        self._data = data or b"\x00" * 32

    @property
    def value(self):
        return unpack_uint256(self._data)

    @value.setter
    def value(self, value):
        self._data = pack_uint256(value)

    def serialize(self):
        return self._data


def unpack_uint128(data):
    assert data is None or (type(data) == bytes and len(data) == 16)

    if data:
        return web3.Web3.toInt(data)
    else:
        return 0


def pack_uint128(value):
    assert value is None or (type(value) == int and value >= 0 and value < 2**128)

    if value:
        data = web3.Web3.toBytes(value)
        return b"\x00" * (16 - len(data)) + data
    else:
        return b"\x00" * 16


class uint128(object):
    def __init__(self, data=None):
        self._data = data or b"\x00" * 16

    @property
    def value(self):
        return unpack_uint128(self._data)

    @value.setter
    def value(self, value):
        self._data = pack_uint128(value)

    def serialize(self):
        return self._data


def unpack_uint64(data):
    assert data is None or (type(data) == bytes and len(data) == 8)

    if data:
        return web3.Web3.toInt(data)
    else:
        return 0


def pack_uint64(value):
    assert value is None or (type(value) == int and value >= 0 and value < 2**64)

    if value:
        data = web3.Web3.toBytes(value)
        return b"\x00" * (8 - len(data)) + data
    else:
        return b"\x00" * 8


class uint64(object):
    def __init__(self, data=None):
        self._data = data or b"\x00" * 8

    @property
    def value(self):
        return unpack_uint64(self._data)

    @value.setter
    def value(self, value):
        self._data = pack_uint64(value)

    def serialize(self):
        return self._data


class address(object):
    def __init__(self, data=None):
        self._data = data or b"\x00" * 20

    @property
    def value(self):
        w2, w1, w0 = struct.unpack(">LQQ", self._data)
        return w2 << 16 + w1 << 8 + w0

    @value.setter
    def value(self, value):
        assert type(value) == int and value >= 0 and value < 2**160
        w0 = value % 2**64
        value = value >> 8
        w1 = value % 2**64
        value = value >> 8
        w2 = value % 2**32
        self._data = struct.pack(">LQQ", w2, w1, w0)

    def serialize(self):
        return self._data


class ConfigurationElement(object):
    """
    CFC configuration element, an abstract "thing" with a unique identifier ("oid",
    which is of type UUID) and that can be user documented using

    * label
    * description
    * tags

    These elements are under user (application) control and not interpreted by
    the CFC backend code (beyond their basic types of string or list of string).

    Configuration elements can be nearly everything in CFC configuration:

    * Users
    * Mrealms
    * Nodes
    * ...
    """

    # oid: uuid.UUID
    oid = None
    """
    Object ID.
    """

    # label: Optional[str]
    label = None
    """
    User label for object (optional free text).
    """

    # description: Optional[str]
    description = None
    """
    User description for object (optional free text)
    """

    # tags: Optional[List[str]]
    tags = None
    """
    User tags for object (optional free list of text portions)
    """

    def __init__(self, oid=None, label=None, description=None, tags=None, _unknown=None):
        self.oid = oid

        self.label = label
        self.description = description
        self.tags = tags

        # private member with unknown/untouched data passing through
        self._unknown = _unknown

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if other.oid != self.oid:
            return False
        if other.label != self.label:
            return False
        if other.description != self.description:
            return False
        if (self.tags and not other.tags) or (not self.tags and other.tags):
            return False
        if other.tags and self.tags:
            if set(other.tags) ^ set(self.tags):
                return False

        # _unknown is not part of comparison

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def copy(self, other, overwrite=False):
        if (not self.oid and other.oid) or overwrite:
            self.oid = other.oid
        if (not self.label and other.label) or overwrite:
            self.label = other.label
        if (not self.description and other.description) or overwrite:
            self.description = other.description
        if (not self.tags and other.tags) or overwrite:
            self.tags = other.tags

        # _unknown is not copied

    def marshal(self):
        assert self.oid is None or isinstance(self.oid, uuid.UUID)
        assert self.label is None or type(self.label) == str
        assert self.description is None or type(self.description) == str
        assert self.tags is None or (type(self.tags) == list and type(tag) == str for tag in self.tags)

        obj = dict()
        obj["oid"] = str(self.oid) if self.oid else None
        obj["label"] = self.label
        obj["description"] = self.description
        obj["tags"] = self.tags

        return obj

    @staticmethod
    def parse(data):
        assert type(data) == dict

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = dict()
        for k in data:
            if k not in ["oid", "label", "description", "tags"]:
                _unknown[k] = data[k]

        oid = None
        if "oid" in data and data["oid"] is not None:
            assert type(data["oid"]) == str
            oid = uuid.UUID(data["oid"])

        label = None
        if "label" in data and data["label"] is not None:
            assert type(data["label"]) == str
            label = data["label"]

        description = None
        if "description" in data and data["description"] is not None:
            assert type(data["description"]) == str
            description = data["description"]

        tags = None
        if "tags" in data and data["tags"] is not None:
            assert type(data["tags"]) == list
            for tag in data["tags"]:
                assert type(tag) == str
            tags = data["tags"]

        obj = ConfigurationElement(oid=oid, label=label, description=description, tags=tags, _unknown=_unknown)

        return obj



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.realmstore._event import Event, Events
from cfxdb.realmstore._publication import Publication, Publications
from cfxdb.realmstore._schema import RealmStore
from cfxdb.realmstore._session import IndexSessionsBySessionId, Session, Sessions

__all__ = (
    "Session",
    "Sessions",
    "IndexSessionsBySessionId",
    "Event",
    "Events",
    "Publication",
    "Publications",
    "RealmStore",
)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.gen.xbrnetwork.AccountLevel import AccountLevel
from cfxdb.gen.xbrnetwork.VerificationStatus import VerificationStatus
from cfxdb.gen.xbrnetwork.VerificationType import VerificationType
from cfxdb.gen.xbrnetwork.WalletType import WalletType

from .account import Account, Accounts, IndexAccountsByEmail, IndexAccountsByUsername, IndexAccountsByWallet
from .schema import Schema
from .userkey import IndexUserKeyByAccount, UserKey, UserKeys
from .vaction import VerifiedAction, VerifiedActions

__all__ = (
    # database schema
    "Schema",
    # enum types
    "AccountLevel",
    "WalletType",
    "VerificationType",
    "VerificationStatus",
    # database tables
    "Account",
    "Accounts",
    "IndexAccountsByUsername",
    "IndexAccountsByEmail",
    "IndexAccountsByWallet",
    "UserKey",
    "UserKeys",
    "IndexUserKeyByAccount",
    "VerifiedAction",
    "VerifiedActions",
)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# This table stores XBRToken._balances state.
class TokenBalance(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TokenBalance()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTokenBalance(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # TokenBalance
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Primary key: XBR token owner address.
    # TokenBalance
    def OwnerAddress(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenBalance
    def OwnerAddressAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenBalance
    def OwnerAddressLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenBalance
    def OwnerAddressIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # XBR token owned.
    # TokenBalance
    def Value(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # TokenBalance
    def ValueAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # TokenBalance
    def ValueLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # TokenBalance
    def ValueIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0


def TokenBalanceStart(builder):
    builder.StartObject(2)


def Start(builder):
    TokenBalanceStart(builder)


def TokenBalanceAddOwnerAddress(builder, ownerAddress):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(ownerAddress), 0)


def AddOwnerAddress(builder, ownerAddress):
    TokenBalanceAddOwnerAddress(builder, ownerAddress)


def TokenBalanceStartOwnerAddressVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOwnerAddressVector(builder, numElems: int) -> int:
    return TokenBalanceStartOwnerAddressVector(builder, numElems)


def TokenBalanceAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0)


def AddValue(builder, value):
    TokenBalanceAddValue(builder, value)


def TokenBalanceStartValueVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartValueVector(builder, numElems: int) -> int:
    return TokenBalanceStartValueVector(builder, numElems)


def TokenBalanceEnd(builder):
    return builder.EndObject()


def End(builder):
    return TokenBalanceEnd(builder)
